///////////////////////////////////////////////////////////////////////////////
//
//    (C) Copyright 1995 - 2000 OSR Open Systems Resources, Inc.
//    All Rights Reserved
//
//    This sofware is supplied for instructional purposes only.
//
//    OSR Open Systems Resources, Inc. (OSR) expressly disclaims any warranty
//    for this software.  THIS SOFTWARE IS PROVIDED  "AS IS" WITHOUT WARRANTY
//    OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING, WITHOUT LIMITATION,
//    THE IMPLIED WARRANTIES OF MECHANTABILITY OR FITNESS FOR A PARTICULAR
//    PURPOSE.  THE ENTIRE RISK ARISING FROM THE USE OF THIS SOFTWARE REMAINS
//    WITH YOU.  OSR's entire liability and your exclusive remedy shall not
//    exceed the price paid for this material.  In no event shall OSR or its
//    suppliers be liable for any damages whatsoever (including, without
//    limitation, damages for loss of business profit, business interruption,
//    loss of business information, or any other pecuniary loss) arising out
//    of the use or inability to use this software, even if OSR has been
//    advised of the possibility of such damages.  Because some states/
//    jurisdictions do not allow the exclusion or limitation of liability for
//    consequential or incidental damages, the above limitation may not apply
//    to you.
//
//    OSR Open Systems Resources, Inc.
//    105 Route 101A Suite 19
//    Amherst, NH 03031  (603) 595-6500 FAX: (603) 595-6503
//    email bugs to: bugs@osr.com
//
//
//    MODULE:
//
//        NothingPnP.C -- Nothing driver V2 pnp functions
//
//    ABSTRACT:
//
//      This file contains the skeleton code for a WDM/Win2K device driver
//      that does absolutely nothing.
//
//    AUTHOR(S):
//
//        OSR Open Systems Resources, Inc.
// 
//    REVISION:   
//
//        This is a WDM compatible version of the original NT V4
//        NOTHING driver.
//
///////////////////////////////////////////////////////////////////////////////

#include <ntddk.h>
#include "nothing.h"


NTSTATUS OsrPnpComplete(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context);

///////////////////////////////////////////////////////////////////////////////
//
//	NothingAddDevice
//
//      We're called here by the PnP Manager when a device instance
//      needs to be created
//
//	INPUTS:
//
//		DriverObj - Address of our DRIVER_OBJECT.
//      PhysicalDeviceObject -- Pointer to underlying PDO
//
//	OUTPUTS:
//
//		None.
//
//	RETURNS:
//
//      NTSTATUS -- result of our attempt to create this instance
//
//  IRQL:
//
//    This routine is called at IRQL_PASSIVE_LEVEL.
//
//	NOTES:
//
//    No hardware access in this routine!!!
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS NothingAddDevice(IN PDRIVER_OBJECT  DriverObject,
                          IN PDEVICE_OBJECT  PhysicalDeviceObject)
{
    PNOTHING_DEVICE_EXT devExt;
    PDEVICE_OBJECT	deviceObject;
    UNICODE_STRING	devName;
    UNICODE_STRING  linkName;
	NTSTATUS		code = STATUS_SUCCESS;		

    DbgPrint(DRIVER_NAME "AddDevice: entered\n");

    //
    // Initialize the UNICODE device name.  This will be the "native NT" name
    // for our device.
    //
    RtlInitUnicodeString(&devName, L"\\Device\\Nothing");

    //
    // Ask the I/O Manager to create the device object and
    // device extension
    //
    code = IoCreateDevice(DriverObject,
                    sizeof(NOTHING_DEVICE_EXT),
                    &devName,
                    FILE_DEVICE_NOTHING,
                    0,       
                    FALSE,
                    &deviceObject);


    if(!NT_SUCCESS(code))  {

#if DBG
        DbgPrint(DRIVER_NAME "IoCreateDevice failed.  Status = 0x%0x\n", code);
#endif
        return(STATUS_UNSUCCESSFUL);
    }    

    //
    // Get a pointer to our device extension
    //
    devExt = (PNOTHING_DEVICE_EXT)deviceObject->DeviceExtension;

    //
    // Save the device object pointer away for future reference...
    //
    devExt->DeviceObject = deviceObject;

    //
    // Next, make the device accessible from user-mode applications.
    // Note that this name can be either the same or different from
    // the native "kernel mode" name of the device object, given above.
    //
    // Note that we either additionally, or alternatively, export a
    // device interface (by calling IoRegisterDeviceInterface()).
    // But, accessing this specific device by name is what we expect
    // to do, so we'll bypass that.
    //
    RtlInitUnicodeString(&linkName, L"\\DosDevices\\Nothing");

    code = IoCreateSymbolicLink(&linkName, &devName);

    if (!NT_SUCCESS(code))
    {

#if DBG
        DbgPrint(DRIVER_NAME "IoCreateSymbolicLink failed.  Status = 0x%x\n", code);
#endif
        IoDeleteDevice(deviceObject);

        //
        // Indicate load failure to the I/O manager; driver image is deleted...
        //
        return(code);
    }


    //
    // Tell the I/O Manger to buffer our reads/writes
    //
    deviceObject->Flags |= DO_BUFFERED_IO;

    //
    // Get a pointer to our device extension
    //
    devExt = (PNOTHING_DEVICE_EXT)deviceObject->DeviceExtension;
    
	//
	//  Clear the Device Initializing bit since the Device Object was created
	//  outside of DriverEntry.
	//
	deviceObject->Flags &= ~DO_DEVICE_INITIALIZING;

    //
    // Save the device object pointer away for future reference
    //
    devExt->DeviceObject = deviceObject;

	//
	// Save the address of the physical device object away for future reference
	//
	devExt->Pdo = PhysicalDeviceObject;

	devExt->DeviceToSendIrpsTo = 
	      IoAttachDeviceToDeviceStack(deviceObject, PhysicalDeviceObject);
    
    if(devExt->DeviceToSendIrpsTo == NULL)  {

#if DBG
        DbgPrint(DRIVER_NAME "IoAttachDeviceToDeviceStack failed to attach to Target Device");
#endif
        //
        // The attempt to attach our device FAILED!
        // (this would be a VERY strange error to get)
        //

        //
        // Clean up the mess
        //
        //
        // Delete the symbolic link that we just created
        //
        IoDeleteSymbolicLink(&linkName);

        //
        // Return our device object
        //
        IoDeleteDevice(deviceObject);

        //
        // Indicate load failure to the I/O manager; driver image is deleted...
        //
        return(STATUS_UNSUCCESSFUL);
    }

#if DBG
    DbgPrint(DRIVER_NAME "AddDevice: done\n");
#endif

	return(STATUS_SUCCESS);
}

static PSTR pnpMinorCodes[] = 
{
    "IRP_MN_START_DEVICE",
    "IRP_MN_QUERY_REMOVE_DEVICE",
    "IRP_MN_REMOVE_DEVICE",
    "IRP_MN_CANCEL_REMOVE_DEVICE",
    "IRP_MN_STOP_DEVICE",
    "IRP_MN_QUERY_STOP_DEVICE",
    "IRP_MN_CANCEL_STOP_DEVICE",
    "IRP_MN_QUERY_DEVICE_RELATIONS",
    "IRP_MN_QUERY_INTERFACE",
    "IRP_MN_QUERY_CAPABILITIES",
    "IRP_MN_QUERY_RESOURCES",
    "IRP_MN_QUERY_RESOURCE_REQUIREMENTS",
    "IRP_MN_QUERY_DEVICE_TEXT",
    "IRP_MN_FILTER_RESOURCE_REQUIREMENTS",
    "***** FUNCTION 0x0e",
    "IRP_MN_READ_CONFIG",
    "IRP_MN_WRITE_CONFIG",
    "IRP_MN_EJECT",
    "IRP_MN_SET_LOCK",
    "IRP_MN_QUERY_ID",
    "IRP_MN_QUERY_PNP_DEVICE_STATE",
    "IRP_MN_QUERY_BUS_INFORMATION",
    "IRP_MN_DEVICE_USAGE_NOTIFICATION",
    "IRP_MN_SURPRISE_REMOVAL",
    "IRP_MN_QUERY_LEGACY_BUS_INFORMATION"
};


///////////////////////////////////////////////////////////////////////////////
//
//	NothingPnP
//
//      This routine handles Plug and Play requests for the nothing
//      driver.
//
//	INPUTS:
//      The usual found at a dispatch entry point
//
//      DeviceObject        Pointer to a device object 
//      IRP                 Pointer to an IRP
//
//
//	OUTPUTS:
//
//		None.
//
//	RETURNS:
//
//      NTSTATUS -- status of request.
//
//  IRQL:
//
//    This routine is called at IRQL_PASSIVE_LEVEL.
//
//	NOTES:
//
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS 
NothingPnp(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    PIO_STACK_LOCATION ioStackLocation = IoGetCurrentIrpStackLocation(Irp);
    PNOTHING_DEVICE_EXT devExt = (PNOTHING_DEVICE_EXT)DeviceObject->DeviceExtension;
    PDEVICE_OBJECT targetDevice;
    KEVENT eventWaitLowerDrivers;
    NTSTATUS code;
    UNICODE_STRING linkName;

#if DBG
    DbgPrint(DRIVER_NAME "PnP: PROCESSING IRP_MJ_PNP\n");

    if(ioStackLocation->MinorFunction <= IRP_MN_QUERY_LEGACY_BUS_INFORMATION )  {

        DbgPrint(DRIVER_NAME "Pnp: *** PNP Minor Function is %s\n", 
                        pnpMinorCodes[ioStackLocation->MinorFunction]);

    } else  {

        DbgPrint(DRIVER_NAME "Pnp: *** PNP Minor Function is 0x%0x\n",
                ioStackLocation->MinorFunction);
    }
#endif

    targetDevice = devExt->DeviceToSendIrpsTo;

    switch(ioStackLocation->MinorFunction)  {
        
        case IRP_MN_START_DEVICE:

#if DBG
            DbgPrint(DRIVER_NAME "Pnp: PROCESSING START_DEVICE\n");
#endif
            KeInitializeEvent(&eventWaitLowerDrivers, NotificationEvent, FALSE);

            //
            // The BUS DRIVER handles this IRP before we do
            //
            IoCopyCurrentIrpStackLocationToNext(Irp);

            //
            // Call OsrPnpComplete() when this IRP is done...
            //
            IoSetCompletionRoutine(Irp,
                                  OsrPnpComplete,
                                  &eventWaitLowerDrivers,
                                  TRUE,
                                  TRUE,
                                  TRUE);

            //
            // Send the IRP to the bus driver.  Let's see what HE
            // thinks.
            //
            code = IoCallDriver(devExt->DeviceToSendIrpsTo, Irp);

            if (STATUS_PENDING == code) {

                KeWaitForSingleObject(&eventWaitLowerDrivers,
                                   Executive,
                                   KernelMode,
                                   FALSE,
                                   NULL);

                code = Irp->IoStatus.Status;
            }

            //
            // Can the bus driver do the start?
            //
            if (!NT_SUCCESS(code)) {

                //
                // The bus driver has declined to start the device.
                // Oh well...
                //
#if DBG
                DbgPrint(DRIVER_NAME "Pnp: IoCallDriver() for START fails! 0x%0x \n", code);
#endif
            } else  {                               

                //
                // If we called IoRegisterDeviceInterface() in our
                // AddDevice routine, we'd call IoSetDeviceInterfaceState()
                // here to set the interface state to enabled.
                //
                
#if DBG
                DbgPrint(DRIVER_NAME "Pnp: IoCallDriver() for START succeeds...\n");
#endif
                code = STATUS_SUCCESS;
            }

            //
            // We must now complete the IRP, since we stopped it in the
            // completetion routine with MORE_PROCESSING_REQUIRED.
            //
            Irp->IoStatus.Status = code;

            Irp->IoStatus.Information = 0;

            IoCompleteRequest(Irp, IO_NO_INCREMENT);

            return(code);            

            break;

        case IRP_MN_QUERY_REMOVE_DEVICE:
        case IRP_MN_CANCEL_REMOVE_DEVICE:
        case IRP_MN_QUERY_STOP_DEVICE:
        case IRP_MN_CANCEL_STOP_DEVICE:
        case IRP_MN_STOP_DEVICE:

            //
            // Set the status that indicates WE'RE good with the request
            //
            Irp->IoStatus.Status = STATUS_SUCCESS;
            Irp->IoStatus.Information = 0;

            break;

        //
        // The only thing we have to take any ACTION on is
        // the request to REMOVE the device
        //
        case IRP_MN_REMOVE_DEVICE:
#if DBG
            DbgPrint(DRIVER_NAME "Pnp: Doing remove\n");
#endif

            //
            // If we called IoRegisterDeviceInterface() in our
            // AddDevice routine, we'd call IoSetDeviceInterfaceState()
            // here to set the interface state to DISabled.
            //

            //
            // Delete the symbolic link that points to this name
            //
            RtlInitUnicodeString(&linkName, L"\\??\\Nothing");

            IoDeleteSymbolicLink(&linkName);

            //
            // Detach from the PDO 
            //
            IoDetachDevice(devExt->DeviceToSendIrpsTo);

            //
            // Return our device object
            //
            IoDeleteDevice(devExt->DeviceObject);

#if DBG
            devExt->DeviceToSendIrpsTo = NULL;

            devExt = NULL;
#endif
            //
            // Set the status that indicate that we're succeeding the
            // remove operation
            //
            Irp->IoStatus.Status = STATUS_SUCCESS;
            Irp->IoStatus.Information = 0;

            break;

        default:
#if DBG
            DbgPrint(DRIVER_NAME "Pnp: Default action\n");
#endif
            break;


    }

    //
    // Setup the I/O Stack Location
    //
    IoSkipCurrentIrpStackLocation(Irp);

    //
    // Send it down, cuz we don't care...
    //
    return(IoCallDriver(targetDevice, Irp));
}


//
// OsrPnpComplete
// 
// This is the completion routine for IRP_MJ_PNP requests
//
NTSTATUS
OsrPnpComplete(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context)
{
    PKEVENT pEvent = (PKEVENT) Context;

    UNREFERENCED_PARAMETER (DeviceObject);

    KeSetEvent(pEvent, 0, FALSE);

    //
    // Take the IRP back so that we can continue using it during
    // the IRP_MN_START_DEVICE dispatch routine.
    // NB: we will have to call IoCompleteRequest
    //
    return STATUS_MORE_PROCESSING_REQUIRED;
}

